import os

class Tasks:
    def __init__(self, carol, task_id=None):
        self.carol = carol

        if task_id is not None:
            self.get_task(task_id=task_id)
        else:
            self.task_id = os.environ.get('LONGTASKID')

    def _set_task_by_json(self, json_task):
        self.task_id = json_task.get('mdmId')
        self.user_id = json_task.get('mdmUserId')
        self.connector_id = json_task.get('mdmConnectorId')
        self.task_ready = json_task.get('mdmTaskReady')
        self.task_processing = json_task.get('mdmTaskProcessing')
        self.task_status = json_task.get('mdmTaskStatus')
        self.task_owner = json_task.get('mdmTaskOwner')
        self.task_progress = json_task.get('mdmTaskProgress')
        self.distribution_value = json_task.get('mdmDistributionValue')
        self.task_priority = json_task.get('mdmTaskPreference')
        self.number_of_steps = json_task.get('mdmNumberOfSteps')
        self.number_of_steps_executed = json_task.get('mdmNumberOfStepsExecuted')
        self.entity_type = json_task.get('mdmEntityType')
        self.created_date = json_task.get('mdmCreated')
        self.start_date = json_task.get('mdmStartDate')
        self.created_user = json_task.get('mdmCreatedUser')
        self.updated_user = json_task.get('mdmUpdatedUser')
        self.last_updated = json_task.get('mdmLastUpdated')
        self.tenant_id = json_task.get('mdmTenantId')
        self.process_after = json_task.get('mdmProcessAfter')
        self.data = json_task.get('mdmData')

    def get_by_task_id_in_env(self):
        task_id = os.getenv('LONGTASKID')
        assert task_id, "The task id has not been set in env by Carol."
        return self.get_task(task_id)

    def get_current_task_id(self):
        return os.getenv('LONGTASKID')

    def set_as_current_task(self):
        os.environ['LONGTASKID'] = self.task_id

    def get_task(self, task_id=None):
        """
        Get Task
        :param task_id: task id
        :return: Task
        """

        if task_id is None:
            task_id = self.task_id
            assert task_id, "Task ID should be set because it has not been set in env by Carol."

        json_task = self.carol.call_api('v1/tasks/{}'.format(task_id), status_forcelist=(500, 502, 503, 504, 524))
        self._set_task_by_json(json_task)
        return self

    def trace(self, log_message, task_id=None):
        self.add_log(log_message, "TRACE", task_id)

    def debug(self, log_message, task_id=None):
        self.add_log(log_message, "DEBUG", task_id)

    def info(self, log_message, task_id=None):
        self.add_log(log_message, "INFO", task_id)

    def warn(self, log_message, task_id=None):
        self.add_log(log_message, "WARN", task_id)

    def error(self, log_message, task_id=None):
        self.add_log(log_message, "ERROR", task_id)

    def add_log(self, log_message, log_level="INFO", task_id=None):
        """
        Add a log
        :param log_message: commonly used tenandId
        :param log_level: options: ERROR, WARN, INFO, DEBUG, TRACE
        :param task_id: it's not necessary if self.id is defined or if we are running from Carol as a batch app
        :return: boolean
        """

        if task_id is None:
            task_id = self.task_id
            assert task_id, "Task ID should be set because it has not been set in env by Carol."

        log = [{
            "mdmTaskId": task_id,
            "mdmLogMessage": log_message,
            "mdmLogLevel": log_level.upper()
        }]
        return self.add_logs(log)

    def add_logs(self, logs, task_id=None):
        """
        Add more than one log

        Args:
            logs: `list`
                list of logs objects [{"task_id":"", "log_message": "", "log_level": ""}]
            task_id `str` default `None`
                    The task ID. it's not necessary if self.task_id is defined

        :return: Task
        """

        if task_id is None:
            task_id = self.task_id
            assert task_id, "Task ID should be set because it has not been set in env by Carol."

        resp = self.carol.call_api('v1/tasks/{}/logs'.format(task_id), data=logs, status_forcelist=(500, 502, 503,
                                                                                                    504, 524))
        return resp['success']

    def get_logs(self, task_id=None):
        """
        Get all logs

        Args:
           task_id `str` default `None`
                The task ID. it's not necessary if self.task_id is defined

        :return:
            list of logs
        """

        if task_id is None:
            task_id = self.task_id
            assert task_id, "Task ID should be set because it has not been set in env by Carol."

        return self.carol.call_api('v1/tasks/{}/logs'.format(task_id), status_forcelist=(500, 502, 503, 504, 524))

    def set_progress(self, progress, progress_data=None, task_id=None):
        """
        Set Task Progress

        Args:
            progress: `int`
                Number relative to progress
            progress_data: 'dict` default `None`
                Json payload to be sent to Carol
            task_id `str` default `None`
                The task ID. It's not necessary if self.task_id is defined

        :return:
            Task response.
        """

        if progress_data is None:
            progress_data = {}
        else:
            assert isinstance(progress_data, dict)

        if task_id is None:
            task_id = self.task_id
            assert task_id, "Task ID should be set because it has not been set in env by Carol."

        return self.carol.call_api('v1/tasks/{}/progress/{}'.format(task_id, progress), data=progress_data,
                                   status_forcelist=(500, 502, 503, 504, 524))

    def cancel(self, task_id=None, force=False):
        """
        Cancel the task

        Args:
            task_id: `str` default `None`
                The task ID. It's not necessary if self.task_id is defined
            force: `bool` default `False`
                Force cancel

        :return:
            boolean
        """

        if task_id is None:
            task_id = self.task_id
            assert task_id, "Task ID should be set because it has not been set in env by Carol."

        params = {"force": force}

        resp = self.carol.call_api('v1/tasks/{}/cancel'.format(task_id), method="POST", params=params)
        return resp['mdmTaskStatus'] == 'CANCELED'

    def fail(self, task_id=None, message=''):
        """
        Fail the task

        Args:
            task_id: `str` default `None`
                The task Id. It's not necessary if self.task_id is defined
            :param message: `str` default ``
                message to log

        :return:
            boolean
        """

        if task_id is None:
            task_id = self.task_id
            assert task_id, "Task ID should be set because it has not been set in env by Carol."

        params = {"message": message}

        resp = self.carol.call_api('v1/tasks/{}/fail'.format(task_id), method="POST", params=params)
        return resp['mdmTaskStatus'] == 'FAILED'



from unittest import TestCase
from ..task import Task
from ..targets import DummyTarget
from .. import PickleTarget
import logging
logger = logging.getLogger(__name__)


class TestTarget(TestCase):
    def test_old_deprecated_style_raises_warning(self):

        class SampleTask(Task):
            TARGET = DummyTarget

        self.assertWarns(DeprecationWarning, SampleTask().output)
        self.assertEqual(SampleTask().output().__class__, DummyTarget)
        # TODO Test target content - luigi execution

    def test_target_type_changing_default_target(self):
        class SampleTask(Task):
            target_type = DummyTarget

        class NoChangesTask(Task):
            pass

        self.assertEqual(SampleTask().output().__class__, DummyTarget)
        self.assertEqual(NoChangesTask().output().__class__, PickleTarget)
        # TODO Test target content - luigi execution



import asyncio
from concurrent.futures import ThreadPoolExecutor

from .miscellaneous import stream_data
import threading


class AtomicCounter:
    """An atomic, thread-safe incrementing counter.

    Args:
        initial: `int` default `0`
            Initial value for the counter.
        total: `int` default `None`
            If exists, the max value that will be reached.

    """

    def __init__(self, initial=0, total=None):
        """Initialize a new atomic counter to given initial value.
        """
        self.value = initial
        self.total = total
        self._lock = threading.Lock()

    def increment(self, num=1):
        """Atomically increment the counter by num (default 1) and return the
        new value.
        """
        with self._lock:
            self.value += num
            return self.value

    def print(self):
        # Guarantee to print only one thread each time.
        with self._lock:
            print(f'{self.value}/{self.total} sent', end='\r')


def send_a(carol, session, url, data_json, extra_headers, content_type, counter):
    """
    Helper function to be used when sending data async.

    Args:
        carol: requests.Session
            Carol object
        session: `requests.Session`
            Session object to handle multiple API calls.
        url: `str`
            end point to be called.
        data_json: `dict`
            The json to be send.
        extra_headers: `dict`
            Extra headers to be used in the API call
        content_type: `dict`
            Content type of the call.
        :return: None
    """
    carol.call_api(url, data=data_json, extra_headers=extra_headers,
                   content_type=content_type, session=session)

    counter.increment(len(data_json))
    counter.print()


async def send_data_asynchronous(carol, data, step_size, url, extra_headers,
                                 content_type, max_workers, compress_gzip):
    """
    Helper function to send data asynchronous.

    Args:
        carol: `pycarol.carol.Carol`.
            Carol object
        data: `pandas.DataFrame` or `list of dict`,
            Data to be sent.
        step_size: 'int'
            Number of records per slice.
        url: 'str'
            API URI
        extra_headers: `dict`
            Extra headers to be used in the API call
        content_type:  `dict`
            Content type of the call.
        max_workers:  `int`
            Max number of workers of the async job
        compress_gzip: 'bool'
            If to compress the data to send
        :return:
    """

    counter = AtomicCounter(total=len(data))
    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        session = carol._retry_session(status_forcelist=[502, 429, 524, 408, 504, 598, 520, 503, 500],
                                       method_whitelist=frozenset(['POST']))
        # Set any session parameters here before calling `send_a`
        loop = asyncio.get_event_loop()
        tasks = [
            loop.run_in_executor(
                executor,
                send_a,
                *(carol, session, url, data_json, extra_headers, content_type, counter)
                # Allows us to pass in multiple arguments to `send_a`
            )
            for data_json, _ in stream_data(data=data,
                                            step_size=step_size,
                                            compress_gzip=compress_gzip)
        ]

        for _ in await asyncio.gather(*tasks):
            pass



"""Functions to inspect bytecodes"""
import importlib
import builtins

from pycarol.pipeline.utils.hash_versioning.hash_versioning import DEBUG_MODE


def get_name_and_code_of_MAKE_FUNCTION(ix, inst, instructions):
    """
    When a MAKE_FUNCTION instruction is found, the name and code pointer can
    be found on the two instructions above.
    Args:
        ix: index of MAKE_FUNCTION instruction
        inst: MAKE_FUNCION instruction
        instructions: list of instructions composing the whole bytecode

    Returns: a dict, whose only key is the function name, and value is function code

    """
    assert inst.opname == "MAKE_FUNCTION"
    assert ix >= 2

    name = instructions[ix - 1].argval.split('.')[-1]  # discard context namespace
    code = instructions[ix - 2].argval

    return {name: code}


def get_name_and_object_of_IMPORT_NAME(ix, inst, instructions):
    """
    When a IMPORT_NAME instruction is found, we can fetch the name of the
    package and its object in this instruction and the next instruction.
    Args:
        ix: index of IMPORT_NAME instruction
        inst: IMPORT_NAME instruction
        instructions: list of instructions composing the whole bytecode

    Returns: a dict, whose only key is the module name, and value is module
    object

    """
    assert inst.opname == "IMPORT_NAME"

    module = instructions[ix].argval
    module_name = instructions[ix+1].argval

    return {module_name: module}


def get_name_of_CALL_FUNCTION(ix, inst, instructions,parent_function):
    """
    When a CALL_FUNCTION instruction is found, the function name is not given
    as a direct argument to this instruction. Instead, the function name can
    be found some instructions above on the bytecode. Between the CALL_FUNCTION
    instruction and the function pointer we found all the function parameters.
    This method implements the logic needed to fetch the function pointer for
    the three kinds of CALL_FUNCTION operations.

    Args:
        ix: index of call function instruction
        inst: call function instruction
        instructions: list of instructions composing the whole bytecode
        parent_function: function object itself

    Returns:
        function_name: the name of the called function or tuple (
        function_name,object) in some cases
    """
    number_of_parameters_in_build_ops = dict(
        BUILD_TUPLE=lambda x: x,
        BUILD_LIST=lambda x: x,
        BUILD_SET=lambda x: x,
        BUILD_MAP=lambda x: 2 * x,
        BUILD_CONST_KEY_MAP=lambda x: x + 1,
        BUILD_STRING=lambda x: x,
        BUILD_TUPLE_UNPACK=lambda x: x,
        BUILD_TUPLE_UNPACK_WITH_CALL=lambda x: x,
        BUILD_LIST_UNPACK=lambda x: x,
        BUILD_SET_UNPACK=lambda x: x,
        BUILD_MAP_UNPACK=lambda x: x,
        BUILD_MAP_UNPACK_WITH_CALL=lambda x: x,
    )

    if "CALL_FUNCTION" == inst.opname:  # it is simple call function instruction
        # for this instruction, we can find the called function some instructions
        # above. we just need to skip backwards the number of arguments
        offset = inst.arg + 1
        # called_function_inst = instructions[ix - offset]
    elif "CALL_FUNCTION_KW" == inst.opname:  # call function op with keyword arguments
        # wrt CALL_FUNCTION there is one additional argument to skip
        offset = inst.arg + 2
        # called_function_inst = instructions[ix - offset]
    elif "CALL_FUNCTION_EX":
        offset = inst.arg + 2
        # Next, we look for BUILD instructions between CALL_FUNCTION_EX instruction
        # and estimated target function position defined by offset. Then, we update
        # offset following each instruction behavior to skip all the parameters
        # and finally find the target function name.
        rev_ind = 0
        while rev_ind < offset:
            rev_ind += 1
            ind_i = ix - rev_ind
            if ind_i < 0:
                raise KeyError("step out of the bytecode while searching CALL_FUNCION_EX target")
            inst_i = instructions[ind_i]
            opname = inst_i.opname
            arg = inst_i.arg
            if opname in number_of_parameters_in_build_ops:  # increase offset
                offset += number_of_parameters_in_build_ops[opname](arg)
    else:
        raise NotImplementedError("instruction {} is not supported".format(inst.opname))

    ix = ix - offset
    called_function_inst = instructions[ix]

    if DEBUG_MODE:
        print(called_function_inst)
        print("offset: ", offset)

    if called_function_inst.opname == 'LOAD_GLOBAL':
        function_name = called_function_inst.argval
    elif called_function_inst.opname == 'LOAD_FAST':
        function_name = called_function_inst.argval
    elif called_function_inst.opname == 'LOAD_ATTR': 
        function_name = get_name_and_object_of_LOAD_ATTR(ix,
                                                         called_function_inst,instructions,parent_function)
        # in this case, function_name is actually (name,object)
    else:
        raise NotImplementedError(
            f"Composed function name not supported."
            f"{called_function_inst.opname}")
    return function_name


def get_function_object_by_name(
        function_name: str,
        enclosing_function, # function or function code
        local_functions: dict
):
    """
    function_name can be found in Local, Enclosed, Global and Builtin Namespace
    Local and Enclosed namespaces cannot be exactly inferred in static analysis.
    For estimating local and enclosed namespaces, we do not consider conditional
    branches(jumps) in the bytecode. All functions definitions encountered are
    evaluated in the order they appear in the bytecode.

    Args:
        function_name: name of called function
        enclosing_function (function object or bytecode): context function in which
        call function happens
        local_functions: dict containing functions defined in enclosing_function

    Returns:
        func: function object
    """

    if function_name in local_functions:  # Local Namespace
        return local_functions[function_name]

    if hasattr(enclosing_function, '__module__'):
        module_namespace = importlib.import_module(enclosing_function.__module__)
        if hasattr(module_namespace, function_name):  # Module (Global) Namespace
            return getattr(module_namespace, function_name)

    if  hasattr(builtins,function_name):#
        return getattr(builtins, function_name)

    raise NotImplementedError("{} was neither found in module {} nor it is builtin.\
    \nEnclosed namespaces are not supported".format(function_name, module_namespace))


def get_name_and_object_of_CALL_FUNCTION(
        ix,inst,instructions,
        parent_function,defined_functions
) -> tuple:
    """
    Wrapper around get_name_of_CALL_FUNCTION and get_function_object_by_name
    Args:
        ix:
        inst:
        instructions:
        parent_function:
        defined_functions:

    Returns:
        tuple function name, function object

    """
    function_name = get_name_of_CALL_FUNCTION(ix,inst,instructions,parent_function)
    if isinstance(function_name,tuple):
        function_name, function_object = function_name
    else:
        function_object = get_function_object_by_name(
            function_name,
            parent_function,
            defined_functions
        )

    from .._utils import is_function, is_builtin, is_code
    assert is_function(function_object) or is_builtin(function_object) or \
           is_code(function_object), f"{function_object} is not a function"

    return function_name, function_object

def get_name_and_object_of_LOAD_ATTR(
        ix,inst,instructions,
        parent_function,
) -> tuple:
    """
    Args:
        ix:
        inst:
        instructions:

    Returns:
        tuple containing function name, function object

    """
    assert inst.opname == 'LOAD_ATTR'

    composed_name = []
    while inst.opname == "LOAD_ATTR":
        composed_name.insert(0,inst.argval)
        ix -= 1
        inst = instructions[ix]
    if inst.opname != "LOAD_GLOBAL":
        raise NotImplementedError(
            f"{inst.opname} not supported with LOAD_ATTR"
        )

    composed_name.insert(0,inst.argval)
    function_name = ".".join(composed_name)

    name = composed_name[0]
    try:
        o = importlib.import_module(name)
    except ModuleNotFoundError:
        m = importlib.import_module(parent_function.__module__)
        o = getattr(m,name)
    for name in composed_name[1:]:
        o = getattr(o,name)
    function_object = o

    return function_name, function_object



import json
import re


class NamedQuery:

    def __init__(self, carol, max_hits=float('inf'), offset=0, page_size=1000, sort_order='ASC', sort_by=None,
                 index_type='MASTER', only_hits=True, fields=None, get_aggs=False,
                 save_results=False, filename=None, print_status=True,
                 ):

        self.named_query_data = []
        self.named_query_dict = {}

        self.carol = carol
        self.max_hits = max_hits
        self.offset = offset
        self.page_size = page_size
        self.sort_order = sort_order
        self.sort_by = sort_by
        self.index_type = index_type
        self.only_hits = only_hits
        self.fields = fields
        self.get_aggs = get_aggs

        self.save_results = save_results

        if self.save_results:
            assert filename, "`save_results=True`, you should specify the filename"
        self.filename = filename
        self.print_status = print_status

        self.param_dict = {}

        self.results = []

        if self.max_hits == float('inf'):
            self._get_all = True
        else:
            self._get_all = False

    def _build_query_params(self):
        self.query_params = {"offset": self.offset, "pageSize": self.page_size, "sortOrder": self.sort_order,
                             "indexType": self.index_type}

        if self.sort_by is not None:
            self.query_params["sortBy"] = self.sort_by

    def _get_param(self, named_query=None):
        assert self.named_query_dict
        self.param_dict = {}
        if named_query is None:
            for key, value in self.named_query_dict.items():
                self.param_dict[key] = re.findall(r'\{\{(.*?)\}\}', json.dumps(value, ensure_ascii=False))

    def get_all(self):

        self._build_query_params()
        self.named_query_data = []
        count = self.offset

        set_param = True
        to_get = float("inf")
        downloaded = 0
        if self.save_results:
            file = open(self.filename, 'w', encoding='utf8')

        url_filter = "v2/named_queries"
        while count < to_get:

            result = self.carol.call_api(url_filter, params=self.query_params)

            if set_param:
                self.total_hits = result["totalHits"]
                if self._get_all:
                    to_get = result["totalHits"]
                elif self.max_hits <= result["totalHits"]:
                    to_get = self.max_hits
                else:
                    to_get = result["totalHits"]
                set_param = False

            count += result['count']
            downloaded += result['count']
            query = result['hits']

            self.named_query_data.extend(query)
            self.named_query_dict.update({i['mdmQueryName']: i for i in query})
            self.query_params['offset'] = count
            if self.print_status:
                print('{}/{}'.format(count, self.total_hits), end='\r')
            if self.save_results:
                file.write(json.dumps(query, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if self.save_results:
            file.close()
        self._get_param()

        return self.named_query_dict

    def by_name(self, named_query):

        self.named_query_data = []
        if self.save_results:
            file = open(self.filename, 'w', encoding='utf8')

        url_filter = "v2/named_queries/name/{}".format(named_query)
        result = self.carol.call_api(url_filter)

        self.named_query_dict.update({result['mdmQueryName']: result})

        if self.save_results:
            file.write(json.dumps(result, ensure_ascii=False))
            file.write('\n')
            file.flush()
            file.close()
        self._get_param()

        return result

    def create_named_query(self, named_query, overwrite=True):

        if isinstance(named_query, dict):
            named_query = [named_query]
        else:
            assert isinstance(named_query, list)

        url_filter = 'v2/named_queries'
        count = 0
        _requet_type = "POST"
        for query in named_query:
            count += 1
            query.pop('mdmId', None)
            query.pop('mdmTenantId', None)

            try:
                old_query = self.by_name(query['mdmQueryName'])
            except Exception as e:
                if 'Not found' in e.args[0]:
                    result = self.carol.call_api(path=url_filter, method=_requet_type, data=query)
            else:
                if overwrite:
                    _requet_type = "PUT"
                    mdm_id = self.named_query_dict[old_query['mdmQueryName']]['mdmId']
                    url_filter = 'v2/named_queries/{}'.format(mdm_id)
                    result = self.carol.call_api(path=url_filter, method=_requet_type, data=query)
                else:
                    print(f"{old_query['mdmQueryName']} "
                          f"already exists and will not be copied, use `overwrite=True` to overwrite")

            print('{}/{} named queries copied'.format(count, len(named_query)), end='\r')



"""
Standard User/Password Auth

"""

import types
import time
from .PwdAuth_cloner import PwdAuthCloner


class PwdAuth:
    """

    Args:

        user: `str`
            Username
        password: `str`
            Password

    """

    def __init__(self, user, password):
        self.user = user
        self.password = password
        self._token = None
        self.carol = None
        self.connector_id = None

    def _set_token(self, data):
        self._token = types.SimpleNamespace()
        self._token.access_token = data['access_token']
        self._token.refresh_token = data['refresh_token']
        self._token.expiration = data['timeIssuedInMillis'] + (
                    data['expires_in'] * 1000)

    def set_connector_id(self, connector_id):
        self.connector_id = connector_id

    def cloner(self):
        return PwdAuthCloner(self)

    def login(self, carol):
        """

        Args:

            carol: pycarol.Carol
                Carol() instance.

        Returns:
            None

        """
        self.carol = carol

        data = {
            'username': self.user,
            'password': self.password,
            'grant_type': 'password',
            'subdomain': carol.domain,
            'connectorId': carol.connector_id,
            'orgSubdomain': carol.organization
        }
        resp = self.carol.call_api('v2/oauth2/token', auth=False, data=data,
                                   content_type='application/x-www-form-urlencoded')

        self._set_token(resp)
        if self.carol.verbose:
            print("Token: {}".format(self._token.access_token))

    def authenticate_request(self, headers):
        headers['Authorization'] = self.get_access_token()

    def _is_token_expired(self):
        if self._token is None:
            return True

        if self._token.expiration == 0:
            return False

        now = time.time() * 1000
        # Adds 1 min buffer
        expiry = self._token.expiration - 60000

        return now > expiry

    def get_access_token(self):
        if self._is_token_expired():
            self._refresh_token()

        return self._token.access_token

    def _refresh_token(self):
        resp = self.carol.call_api('v2/oauth2/token', auth=False, data={
            'grant_type': 'refresh_token',
            'refresh_token': self._token.refresh_token
        }, content_type='application/x-www-form-urlencoded')

        self._set_token(resp)

    def switch_context(self, env_id):
        """
        Switch context to an environment within the same organization.

        Args:

            env_id: `str`
                environment id to switch context to.

        Returns:

            None

        """

        path = f'v2/oauth2/switchTenantContext/{env_id}'
        resp = self.carol.call_api(method='POST', path=path)

        self._set_token(resp)

    def switch_org_context(self, org_id):
        """
        Go to the organization context or switch organization.

        Args:

            org_id: `str`
                organization id.

        Returns:

            None

        """

        path = f'v2/oauth2/switchOrgContext/{org_id}'
        resp = self.carol.call_api(method='POST', path=path)

        self._set_token(resp)



"""
This module hosts the class Pipe, which is used to interact with an
instantiated pipeline.
In each DS app, we expect to find a file named pipeline.py in folder app. In
this file, there will be an object named pipeline, from the class Pipe. In
order to initialize a Pipe object, we need to define a list of output tasks
and a dictionary containing the pipeline parameters. These parameters are
similar to the parameters one pass to luigi.build, however, in this case,
we use a list of non instantiated tasks.
In file test_tools.py there are some basic tests for this module.
"""
from pycarol.pipeline.tools._tools import Pipe


from setuptools import setup, find_packages

import codecs
import os
import re
import sys

here = os.path.abspath(os.path.dirname(__file__))


min_requires = [
    'google-auth-httplib2',
    'google-auth',
    'google-cloud-core>=1.4.1',
    'google-cloud-storage',
    'requests',
    'tqdm',
    'urllib3',
    'deprecated',
    'python-dotenv',
    'gcsfs>=0.3.0,<0.7',
    "retry"
]

extras_require = {
    "dataframe": min_requires + ['pandas>=0.23.4,!=1.0.4', 'numpy>=1.16.3', 'joblib>=0.11', 'pyarrow>=0.15.1,<1.0.0',],
    "pipeline": min_requires + ['luigi', 'papermill', 'pandas>=0.23.4,!=1.0.4', 'numpy>=1.16.3', 'joblib>=0.11',
                                'pyarrow>=0.15.1,<1.0.0',],
    "onlineapp": min_requires + ['flask>=1.0.2', 'redis'],
    "dask": min_requires + ['dask[complete]'],
    "dev": min_requires + ['pytest', 'bumpversion', "sphinx-rtd-theme", "sphinx"],
}
extras_require["complete"] = sorted(
    {v for req in extras_require.values() for v in req}
)


def read(*parts):
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

readme_note = """\
.. note::
   For the latest source, discussion, etc, please visit the
   `GitHub repository <https://github.com/totvslabs/pyCarol>`_\n\n
"""

with open('README.rst') as fobj:
    long_description = readme_note + fobj.read()

def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        version_file,
        re.M,
    )
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")

setup(
    name='pycarol',
    setup_requires=["wheel"],
    packages=find_packages(exclude=['docs', 'doc']),
    version=find_version("pycarol", "__init__.py"),
    license='TOTVS',
    description='Carol Python API and Tools',
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author='TotvsLabs',
    maintainer='TOTVS Labs',
    author_email='ops@totvslabs.com',
    url='https://github.com/totvslabs/pyCarol',
    keywords=['Totvs', 'Carol.ai', 'AI'],
    install_requires=min_requires,
    extras_require=extras_require,
    classifiers=[
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Development Status :: 5 - Production/Stable',
        # Define that your audience are developers
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        "Operating System :: OS Independent",
    ],
)



"""
The main Carol's storage is called CDS (Carol Data Storage). Any data received or created in Carol is sent to CDS.
Inside CDS one can have three kinds of data. Data coming from the
Staging Area, data processed and mapped to a DataModel (Golden Record), and any other file that the user could send.
The `pycarol.cds.CDSStaging` and
the `pycarol.cds.CDSGolden` classes are used to manipulate the data inside the first two cases.

"""
from .connectors import Connectors
from .data_models import DataModel
from .utils.deprecation_msgs import _deprecation_msgs
import warnings

_MACHINE_FLAVORS = [
    'n1-standard-1',
    'n1-standard-2',
    'n1-standard-4',
    'n1-standard-8',
    'n1-standard-16',
    'n1-highmem-2',
    'n1-highmem-4',
    'n1-highmem-8',
    'n1-highmem-16',
    'n2-standard-2',
    'n2-standard-4',
    'n2-standard-8',
    'n2-standard-16',
    'n2-highmem-2',
    'n2-highmem-4',
    'n2-highmem-8',
    'n2-highmem-16',
    'n2-highmem-32',
    'n2-highmem-64',
]

def check_worker_type(worker_type):
    if worker_type not in _MACHINE_FLAVORS and worker_type is not None:
        warnings.warn(
        f'worker_type {worker_type} is unknown. It might not work. Use one of the known machines {_MACHINE_FLAVORS}',
        Warning, stacklevel=3
    )

class CDSStaging:
    """
    Class to handle all CDS Staging iterations.

    """

    def __init__(self, carol):
        self.carol = carol

    def process_data(
            self, staging_name, connector_id=None, connector_name=None,
            worker_type=None, max_number_workers=-1, number_shards=-1, num_records=-1,
            delete_target_folder=False, enable_realtime=None, delete_realtime_records=False,
            send_realtime=None, file_pattern='*', filter_query=None, skip_consolidation=False,
            force_dataflow=False, recursive_processing=True, dm_name=None, auto_scaling=True,
    ):
        """
        Process CDS staging data.

        Args:

            staging_name: `str`,
                Staging name.
            connector_id: `str`, default `None`
                Connector id.
            connector_name: `str`, default `None`
                Connector name.
            worker_type: `str`, default None
                Machine flavor to be used. If `None` Carol will decide the machine to use.
            max_number_workers: `int`, default `-1`
                Max number of workers to be used during the process. '-1' means all the available.
            number_shards: `int`, default `-1`
                Number of shards.
            num_records: `int`, default `-1`
                Number of records to be processed. '-1' means all the records.
            delete_target_folder: `bool`, default `False`
                If delete the previous processed records.
            enable_realtime: `bool`, default `False`
                DEPRECATED. Removed from Carol.
                Enable this staging table to send the processed data to realtime layer.
            delete_realtime_records: `bool`, default `False`
                Delete previous processed data in realtime.
            send_realtime: `bool`, default `None`
                Send the processed data to realtime layer.
            file_pattern: `str`, default `*`
                File pattern of the files in CDS to be processed. The pattern in  `YYYY-MM-DDTHH_mm_ss*.parquet`.
                One can use this to filter data in CDS received in a given date.
            filter_query: `dict`, default `None`
                Query to be used to filter the data to be processed.
            skip_consolidation: `bool` default `False
                If consolidation process should be skipped
            force_dataflow: `bool`  default `False`
                If Dataflow job should be spinned even for small datasets
                (by default, small datasets are processed directly inside Carol)
            recursive_processing: `bool`  default `True`
                If processing should be chained/recursed in target entities. e.g., If a staging has 3 ETLs and each ETL
                maps to a data model. If we process this staging it will trigger the whole tree to be processed.
            dm_name: `str` default `None`
                If not None, it will reprocess the rejected records from the selected staging table.
            auto_scaling: `bool` default `True`
                Use auto scaling. It `False` Carol will use max_number_workers for the whole process.

        :return: dict
            Task definition.

        """

        check_worker_type(worker_type=worker_type)

        if enable_realtime is not None:
            _deprecation_msgs(
                "`enable_realtime` is deprecated and it is not used in Carol. ")

        filter_query = filter_query if filter_query else {}

        if connector_name:
            connector_id = Connectors(self.carol).get_by_name(
                connector_name)['mdmId']
        else:
            if connector_id is None:
                raise ValueError(
                    'connector_id or connector_name should be set.')

        query_params = {
            "connectorId": connector_id, "stagingType": staging_name, "workerType": worker_type,
            "maxNumberOfWorkers": max_number_workers, "numberOfShards": number_shards,
            "numRecords": num_records,
            "deleteTargetFolder": delete_target_folder,
            "deleteRealtimeRecords": delete_realtime_records,
            "sendToRealtime": send_realtime, "filePattern": file_pattern,
            "skipConsolidation": skip_consolidation,
            "forceDataflow": force_dataflow,
            "recursiveProcessing": recursive_processing,
            "rejectedType": dm_name, "autoScaling": auto_scaling,
        }

        return self.carol.call_api(path='v1/cds/staging/processData', method='POST', params=query_params,
                                   data=filter_query)

    def sync_data(
            self, staging_name, connector_id=None, connector_name=None, num_records=-1,
            delete_realtime_records=False, enable_realtime=None,
            file_pattern='*', filter_query=None, force_dataflow=False, records_percentage=100
    ):
        """
        Sync data to realtime layer.

        Args:

            staging_name: `str`,
                Staging name.
            connector_id: `str`, default `None`
                Connector id.
            connector_name: `str`, default `None`
                Connector name.
            num_records: `int`, default `-1`
                Number of records to be processed. '-1' means all the records.
            enable_realtime: `bool`, default `False`
                DEPRECATED. Removed from Carol.
                Enable this staging table to send the processed data to realtime layer.
            delete_realtime_records: `bool`, default `False`
                Delete previous processed data in realtime.
            file_pattern: `str`, default `*`
                File pattern of the files in CDS to be processed. The pattern in  `YYYY-MM-DDTHH_mm_ss*.parquet`.
                One can use this to filter data in CDS received in a given date.
            filter_query: `dict`, default `None`
                Query to be used to filter the data to be processed.
            force_dataflow: `bool`  default `False`
                If Dataflow job should be spinned even for small datasets
                (by default, small datasets are processed directly inside Carol)
            records_percentage" `int` default `100`
                The percentage of records (0-100) to import

        :return: None
        """

        if enable_realtime is not None:
            _deprecation_msgs(
                "`enable_realtime` is deprecated and it is not used in Carol. ")

        filter_query = filter_query if filter_query else {}

        if connector_name:
            connector_id = Connectors(self.carol).get_by_name(
                connector_name)['mdmId']
        else:
            if connector_id is None:
                raise ValueError(
                    'connector_id or connector_name should be set.')

        query_params = {
            "connectorId": connector_id, "stagingType": staging_name,
            "numRecords": num_records,
            "clearStagingRealtime": delete_realtime_records, "filePattern": file_pattern,
            "forceDataflow": force_dataflow, "recordsPercentage": records_percentage,
        }

        return self.carol.call_api(path='v1/cds/staging/fetchData', method='POST', params=query_params,
                                   data=filter_query)

    def consolidate(
            self, staging_name, connector_id=None, connector_name=None,
            worker_type=None, max_number_workers=-1, number_shards=-1, force_dataflow=False,
            rehash_ids=False, file_pattern="*.parquet", compute_transformations=False,
            auto_scaling=True, file_size_limit=-1,
    ):
        """
        Process staging CDS data.

        Args:

            staging_name: `str`,
                Staging name.
            connector_id: `str`, default `None`
                Connector id.
            connector_name: `str`, default `None`
                Connector name.
            worker_type: `str`, default `None`
                Machine flavor to be used. If `None` Carol will decide the machine to use.
            max_number_workers: `int`, default `-1`
                Max number of workers to be used during the process. '-1' means all the available.
            number_shards: `int`, default `-1`
                Number of shards.
            force_dataflow: `bool`  default `False`
                If Dataflow job should be spinned even for small datasets
                (by default, small datasets are processed directly inside Carol)
            rehash_ids" `bool` default `False`
                If all ids should be regenerated from the crosswalk
            file_pattern: `str`, default `*.parquet`
                File pattern of the files in CDS to be consolidated. The pattern is `YYYY-MM-DDTHH_mm_ss*.parquet`.
                One can use this to filter data in CDS received in a given date.
            compute_transformations: `bool` default False
                If staging transformations are defined, this will apply the transformations during the consolidate.
            auto_scaling: `bool` default `True`
                Use auto scaling. It `False` Carol will use max_number_workers for the whole process.
            file_size_limit: `int` default `-1`
                Ignore files larger than "file_size_limit" bytes during consolidation. 


        :return: `dict`
            Task created in Carol. 

        """

        check_worker_type(worker_type=worker_type)

        if connector_name:
            connector_id = Connectors(self.carol).get_by_name(
                connector_name)['mdmId']
        else:
            if connector_id is None:
                raise ValueError(
                    'connector_id or connector_name should be set.')

        query_params = {
            "connectorId": connector_id, "stagingType": staging_name,
            "workerType": worker_type, "maxNumberOfWorkers": max_number_workers,
            "numberOfShards": number_shards, "filePattern": file_pattern,
            "rehashIds": rehash_ids, "forceDataflow": force_dataflow,
            "computeTransformations": compute_transformations,
            "autoScaling": auto_scaling, "fileSizeLimitBytes": file_size_limit,
        }

        return self.carol.call_api(path='v1/cds/staging/consolidate', method='POST', params=query_params)

    def delete(self, staging_name, connector_id=None, connector_name=None):
        """
        Delete all CDS staging data.

        Args:

            staging_name: `str`,
                Staging name.
            connector_id: `str`, default `None`
                Connector id.
            connector_name: `str`, default `None`
                Connector name

        :return: None

        """

        if connector_name:
            connector_id = Connectors(self.carol).get_by_name(
                connector_name)['mdmId']
        else:
            if connector_id is None:
                raise ValueError(
                    'connector_id or connector_name should be set.')

        query_params = {"connectorId": connector_id,
                        "stagingType": staging_name}

        return self.carol.call_api(path='v1/cds/staging/clearData', method='POST', params=query_params)

    def count(self, staging_name, connector_id=None, connector_name=None):
        """

        Count number of messages in CDS.

        Args:

            staging_name: `str`,
                Staging name.
            connector_id: `str`, default `None`
                Connector id.
            connector_name: `str`, default `None`
                Connector name

        :return: `int`
            Count

        """

        if connector_name:
            connector_id = Connectors(self.carol).get_by_name(
                connector_name)['mdmId']
        else:
            if connector_id is None:
                raise ValueError(
                    'connector_id or connector_name should be set.')

        query_params = {"connectorId": connector_id,
                        "stagingType": staging_name}
        return self.carol.call_api(path='v1/cds/staging/fetchCount', method='POST', params=query_params).get('count')


class CDSGolden:
    """
    Class to handle all CDS Staging iterations.

    Args:

        carol: 'pycarol.Carol`
            Carol() instance.

    """

    def __init__(self, carol):
        """
        Args:
            carol:
        """

        self.carol = carol

    def sync_data(self, dm_name, dm_id=None, num_records=-1, file_pattern='*', filter_query=None,
                  skip_consolidation=False, force_dataflow=False, records_percentage=100, worker_type=None,
                  max_number_workers=-1, clear_golden_realtime=False,
                  ):
        """
        Sync data to realtime layer.

        Args:

            dm_name: `str`,
                Data model name.
            dm_id: `str`, default `None`
                Data model id.
            num_records: `int`, default `-1`
                Number of records to be processed. '-1' means all the records.
            file_pattern: `str`, default `*`
                File pattern of the files in CDS to be processed. The pattern in  `YYYY-MM-DDTHH_mm_ss*.parquet`.
                One can use this to filter data in CDS received in a given date.
            filter_query: `dict`, default `None`
                Query to be used to filter the data to be processed.
            skip_consolidation: `bool` default `False
                If consolidation process should be skipped
            force_dataflow: `bool`  default `False`
                If Dataflow job should be spinned even for small datasets
                (by default, small datasets are processed directly inside Carol)
            records_percentage" `int` default `100`
                The percentage of records (0-100) to import
            worker_type: `str`, default `None`
                Machine flavor to be used. If `None` Carol will decide the machine to use.
            max_number_workers: `int`, default `-1`
                Max number of workers to be used during the process. '-1' means all the available.
            clear_golden_realtime: `bool`, default `False`
                If the records on realtime should be deleted first

        :return: None

        """

        filter_query = filter_query if filter_query else {}

        if dm_name:
            dm_id = DataModel(self.carol).get_by_name(dm_name)['mdmId']
        else:
            if dm_id is None:
                raise ValueError('dm_name or dm_id should be set.')

        query_params = {
            "entityTemplateId": dm_id, "numRecords": num_records, "filePattern": file_pattern,
            "skipConsolidation": skip_consolidation, "forceDataflow": force_dataflow,
            "clearGoldenRealtime": clear_golden_realtime, "maxNumberOfWorkers": max_number_workers,
            "workerType": worker_type, "recordsPercentage": records_percentage,

        }

        return self.carol.call_api(path='v1/cds/golden/fetchData', method='POST', params=query_params,
                                   data=filter_query)

    def delete_rejected(self, dm_name=None, dm_id=None, connector_name=None, connector_id=None, staging_name=None):
        """
        Delete CDS DataModel rejected data

        Args:
            dm_name: `str`,
                Data Model name.
            dm_id: `str`, default `None`
                Data Model id.
            connector_name: `str`, default `None`
                Connector name. Used if delete only records from a given connector/staging table
            connector_id:  `str`, default `None`
                Connector id. Used if delete only records from a given connector/staging table
            staging_name:  `str`, default `None`
                Staging name. Used if delete only records from a given connector/staging table

        Returns: dict
            Carol task created.


        """

        if dm_name:
            dm_id = DataModel(self.carol).get_by_name(dm_name)['mdmId']
        else:
            if dm_id is None:
                raise ValueError('dm_name or dm_id should be set.')

        if connector_name:
            connector_id = Connectors(self.carol).get_by_name(
                connector_name)['mdmId']

        if (staging_name is not None) and (connector_id is None):
            raise ValueError(
                'connector_id or connector_name must be set when using staging_name.')

        params = {'entityTemplateId': dm_id,
                  'connectorId': connector_id,
                  'stagingTable': staging_name}

        return self.carol.call_api("v2/cds/rejected/clearData", method='POST',
                                   params=params)

    def delete(self, dm_name=None, dm_id=None, ):
        """
        Delete all CDS data model data.

        Args:

            dm_name: `str`,
                Data Model name.
            dm_id: `str`, default `None`
                Data Model id.

        :return: None
        """

        if dm_name:
            dm_id = DataModel(self.carol).get_by_name(dm_name)['mdmId']
        else:
            if dm_id is None:
                raise ValueError('dm_name or dm_id should be set.')

        query_params = {"entityTemplateId": dm_id}

        return self.carol.call_api(path='v1/cds/golden/clearData', method='POST', params=query_params)

    def count(self, dm_name=None, dm_id=None):
        """
        Count number of messages in CDS.

        Args:

            dm_name: `str`,
                Data Model name.
            dm_id: `str`, default `None`
                Data Model id.

        :return: `int`
            Count
        """

        if dm_name:
            dm_id = DataModel(self.carol).get_by_name(dm_name)['mdmId']
        else:
            if dm_id is None:
                raise ValueError('dm_name or dm_id should be set.')

        query_params = {"entityTemplateId": dm_id}
        return self.carol.call_api(path='v1/cds/golden/fetchCount', method='POST', params=query_params).get('count')

    def consolidate(
        self, dm_name=None, dm_id=None,
        worker_type=None, max_number_workers=-1, number_shards=-1, force_dataflow=False,
        ignore_merge=False, file_pattern="*.parquet",  auto_scaling=True,
        file_size_limit=-1,
    ):
        """

        Process staging CDS data.

        Args:

            dm_name: `str`,
                Data Model name.
            dm_id: `str`, default `None`
                Data Model id.
            worker_type: `str`, default `None`
                Machine flavor to be used. If `None` Carol will decide the machine to use.
            max_number_workers: `int`, default `-1`
                Max number of workers to be used during the process. '-1' means all the available.
            number_shards: `int`, default `-1`
                Number of shards.
            ignore_merge: `bool` default `False
                If merge rules should be ignored when consolidating the records
            force_dataflow: `bool`  default `False`
                If Dataflow job should be spinned even for small datasets
                (by default, small datasets are processed directly inside Carol)
            file_pattern: `str`, default `*.parquet`
                File pattern of the files in CDS to be consolidated. The pattern is `YYYY-MM-DDTHH_mm_ss*.parquet`.
                One can use this to filter data in CDS received in a given date.
            auto_scaling: `bool` default `True`
                Use auto scaling. It `False` Carol will use max_number_workers for the whole process.
            file_size_limit: `int` default `-1`
                Ignore files larger than "file_size_limit" bytes during consolidation. 

        :return: `dict`
            Task created in Carol. 
        """

        if dm_name:
            dm_id = DataModel(self.carol).get_by_name(dm_name)['mdmId']
        else:
            if dm_id is None:
                raise ValueError('dm_name or dm_id should be set.')

        check_worker_type(worker_type=worker_type)

        query_params = {
            "entityTemplateId": dm_id,
            "workerType": worker_type, "maxNumberOfWorkers": max_number_workers,
            "numberOfShards": number_shards,
            "forceDataflow": force_dataflow, "ignoreMerge": ignore_merge, "filePattern": file_pattern,
            "autoScaling": auto_scaling, "fileSizeLimitBytes": file_size_limit,
        }

        return self.carol.call_api(path='v1/cds/golden/consolidate', method='POST', params=query_params)



import luigi
import os
from ..targets import DummyTarget, PickleTarget
from collections import namedtuple
from unittest.mock import patch, MagicMock, PropertyMock
from contextlib import ExitStack
import shutil
import logging
from luigi.execution_summary import LuigiStatusCode
from pycarol.pipeline import Task
logger = logging.getLogger(__name__)

""" 
# Task Execution
    When testing a Task execution, there a couple of things that could be tested:
        - Execution success
        - Requires
        - Execution
# Mocks
# Special Words when defining Test Cases:
    * Test
    Every test case must start with the 'Test' keyword as default from python.unittest.
    * Flow
    If your test case involves Luigi Extension related tasks, use 'Flow' keyword. E.g. TestFlowMyTest. This way, pipeline 
    extension updates will be able to be checked.
"""


def task_execution_debug(task, parameters=None, worker_scheduler_factory=None, **env_params):
    """ Execute a pipeline pipeline
    :param task:
    :param parameters: dict
    :return: instance of class TaskOutput:
            success: bool,
            worker: worker object,
            task: instance task
            task_history: pipeline's worker _add_task_history
            history_has(task, status, ignore_parameters=False): whether task history has or not that task and status.
                Obs. If you want to ignore parameter, make sure task is a class and not an instance.
    """
    if parameters is None:
        parameters = {}

    if "no_lock" not in env_params:
        env_params["no_lock"] = True

    if "local_scheduler" not in env_params:
        env_params["local_scheduler"] = True

    out = dict()
    # TODO Get only parameters that are used in task_instance. Similar to self.clone
    task_instance = task(**parameters)
    out['task'] = task_instance
    exec_out = luigi.interface._schedule_and_run([task_instance], worker_scheduler_factory,
                                                 override_defaults=env_params)
    # TODO: Check luigi version
    # if luigi.__version__
    out.update({'success': exec_out.status==LuigiStatusCode.SUCCESS})
    task_history = exec_out.worker._add_task_history
    out.update({'task_history': task_history})

    def history_has(task, status, ignore_parameters=True):
        if not ignore_parameters:
            for t, s, _ in task_history:
                if task == t and status == s:
                    return True
            return False
        else:
            for t, s, _ in task_history:
                if task.__name__ == t.__class__.__name__ and status == s:
                    return True
            return False

    out.update({'history_has': history_has})

    # TODO Get execution stacktrace
    return namedtuple("TaskOutput", out.keys())(*out.values())


def pipeline_test(cls):
    """ Mock pipeline Task to have TARGET_DIR inside test directory and erase target files before each test
    """
    new_target = f'luigi_targets/test/{cls.__module__}/{cls.__name__}'
    class_setUp = cls.setUp

    def mocked_setUp(self):
        patcher = patch('pycarol.pipeline.Task.TARGET_DIR', new_callable=PropertyMock, return_value=new_target)
        self.addCleanup(patcher.stop)
        self.mock_target = patcher.start()
        if os.path.isdir(new_target):
            shutil.rmtree(new_target)
        return class_setUp(self)

    cls.setUp = mocked_setUp

    return cls


class mock_task:
    """ Define a task as executed and default return from a specific Task
    This mock will work for all Tasks. If the user wants to mock diferently with different parameters, must specify
    task_parameters.

    Dict Parameters:
        mock_task
        task_output or target_filename
        limit_size:
        target_path:
    """

    def __init__(self, *mock_tasks):
        """
        :param mock_tasks: list of tasks
        """
        self.mock_tasks = mock_tasks

    def __call__(self, exec_func):
        mock_tasks = self.mock_tasks

        def patched_func(self, *args, **kwargs):
            with ExitStack() as stack:
                patches = []  # TODO Initialize only if it does not exist. Necessary for wrapper of decorators
                args = [arg for arg in args if not isinstance(arg, MagicMock) and not isinstance(arg, PropertyMock)]
                for dic in mock_tasks:
                    task = dic['mock_task']
                    if 'task_output' in dic:
                        task_output = dic['task_output']
                    elif 'target_filename' in dic:
                        target_filename = dic['target_filename']
                    else:
                        raise ValueError('Mocked Task must have a predefined task_output or target_filename')

                    if 'limit_size' in dic:
                        task_output = task_output[0:dic['limit_size']]

                    if 'task_parameters' in dic:
                        # TODO handle cases of user having same task with different parameters
                        pass
                    else:
                        if 'task_output' in dic:
                            out_target = DummyTarget(fixed_output=task_output)

                        if 'target_filename' in dic:
                            if 'target_class' in dic:
                                TARGET = dic['target_class']
                            else:
                                TARGET = PickleTarget
                            out_target = TARGET(task, path=target_filename, is_tmp=True)
                            out_target.remove = lambda: None  # Use this to avoid having the file removed

                        if 'target_params' in dic:
                            for param_name, param in dic['target_params'].items():
                                setattr(out_target, param_name, param)

                        patches.append([
                            stack.enter_context(
                                patch.object(task, 'output', return_value=out_target)),
                            stack.enter_context(
                                patch.object(task, 'complete', return_value=True))])
                exec_func(self, *args, **kwargs)

        return patched_func


class mock_task_wrapper:
    """ Define a task as executed and default return from a specific Task, but still executes task's requires
    This mock will work for all Tasks. If the user wants to mock diferently with different parameters, must specify
    task_parameters.
    -- Possible improvements:
    An easier way to define a task as completed or not, without using mocks, would be to get the task's output name
    and create that output using 'task_output' as a pickle.
    """

    def __init__(self, *mock_tasks):
        """
        :param mock_tasks: list of tasks
        """
        self.mock_tasks = mock_tasks

    def __call__(self, exec_func):
        def patched_func(*args, **kwargs):
            with ExitStack() as stack:
                patches = []
                for dic in self.mock_tasks:
                    task = dic['mock_task']
                    if 'task_parameters' in dic:
                        # TODO handle cases of user having same task with different parameters
                        pass
                    if 'task_output' in dic:
                        task_output = dic['task_output']
                        out_target = DummyTarget(is_tmp=True)

                        def new_load():
                            return task_output

                        out_target.load = new_load
                        patches.append(stack.enter_context(
                            patch.object(task, 'output', return_value=out_target)))
                    patches.append(
                        stack.enter_context(
                            patch.object(task, 'complete', side_effect=luigi.WrapperTask.complete)))
                exec_func(*args, **kwargs)

        return patched_func


class TaskA(Task):
    def easy_run(self, inputs):
        return True


def test_task_execution_debug():
    out = task_execution_debug(TaskA)
    assert out.success
    assert out.history_has(TaskA, 'DONE')


""" Tests for deprecated features inside pyCarol

"""
import warnings
import pytest


def test_luigi_extension_deprecated():
    with warnings.catch_warnings(record=True) as w:
        # Cause all warnings to always be triggered.
        warnings.simplefilter("default", category=DeprecationWarning)
        import pycarol.luigi_extension
        assert "luigi_extension is deprecated" in str(w[0].message)



from importlib import import_module
from flask import Flask, Response
from flask import request as flask_request
import numpy as np
import os
import json

from .health_check_online import HealthCheckOnline
from .online_request import OnlineRequest
from werkzeug.local import Local, LocalProxy


class OnlineApi():
    """ Class to execute Online API locally
    Use the command line "gunicorn run_me:flask" to execute the code below as an API.

    Usage:
        from pycarol.app.online_api import OnlineApi
        from pycarol.app.online import Online

        online = Online()

        @online.route("sample_endpoint")
        def sample1_function(args):
            return str(args)

        # "run_me" is the filename
        flask = OnlineApi('run_me').get_api()
    """
    def __init__(self, file_name=None, file_path='', domain=None, app_name=None, app_version=None, online_name=None):
        self.file_path = file_path
        self.imported_module = None
        self.endpoints = dict()
        self.logs = []

        if domain:
            self.domain = domain
        else:
            self.domain = os.environ.get('CAROLDOMAIN', 'domain')

        if app_version:
            self.app_version = app_version
        else:
            self.app_version = os.environ.get('CAROLAPPVERSION', 'version')

        if app_name:
            self.app_name = app_name
        else:
            self.app_name = os.environ.get('CAROLAPPNAME', 'name')

        if online_name:
            self.online_name = online_name
        else:
            self.online_name = os.environ.get('CAROLONLINENAME', 'name')

        if not file_name:
            file_name = os.environ.get('ALGORITHMNAME')

        if file_name:
            self.module_name = os.path.splitext(file_name)[0]
        else:
            self.module_name = ''
            self._log_append('The file name should be defined by parameter ou environment variable')

        self._dynamic_import()
        self._load_endpoints()
        self._health_check_carol()

    def _log_append(self, msg):
        self.logs.append(msg)
        print(msg)

    def _dynamic_import(self):
        try:
            self.imported_module = import_module(f'{self.file_path}{self.module_name}')
        except Exception as e:
            self._log_append(f'Problem when importing file. Module: {self.module_name}. Error: {str(e)}')

    def _load_endpoints(self):
        try:
            if self.imported_module:
                for i in dir(self.imported_module):
                    if type(getattr(self.imported_module, i)).__name__ == 'Online':
                        online = getattr(self.imported_module, i)
                        self.endpoints.update(online.get_endpoints())
        except Exception as e:
            self._log_append(f'Problem when trying to load module. Module: {self.module_name}. Error: {str(e)}')

    def _health_check_carol(self):
        healthCheckOnline = HealthCheckOnline(self.logs)
        healthCheckOnline.send_status_carol()

    def get_api(self, debug=False):
        flask = Flask(__name__)

        @flask.route('/', methods=['GET', 'POST'])
        def base():
            return 'Running! Use /api/(endpoint) to access the app api'

        @flask.route(f'/api/<api_path>', methods=['GET', 'POST'])
        def app(api_path):
            try:
                api = self.endpoints[str(api_path)]
            except:
                return f'Endpoint {api_path} not found'

            local.request = OnlineRequest(values=flask_request.values, json=flask_request.json)
            r = api()
            if isinstance(r, np.ndarray):
                r = r.tolist()
            if isinstance(r, tuple):
                resp, code = r
                return Response(json.dumps(resp), status=code, mimetype='application/json')
            return json.dumps(r)

        @flask.route(f'/statusz')
        def app_statusz():
            return 'ok'

        @flask.route(f'/healthz')
        def app_healthz():
            return 'ok'

        @flask.route(f'/logs')
        def app_logs():
            return str(self.logs)

        flask.debug = debug
        return flask

    def run(self, debug=False):
        flask = self.get_api(debug)
        flask.run()
        return flask


local = Local()
request = LocalProxy(local, 'request')


import json
import gzip, io, zipfile, os
from collections import defaultdict
from pathlib import Path
_FILE_MARKER = '<files>'


def drop_duplicated_parquet_dask(d, untie_field='mdmCounterForEntity'):
    """
    Merge updates and delete records from the parquet files in CDS.

    Args:
        d: dask DataFrame
        untie_field: str
            Field to be used to untie records with the same `mdmId`. 

    Returns:
        dask DataFrame

    """

    if untie_field not in d.columns:
        #Use the standard one. 
        untie_field = 'mdmCounterForEntity'


    d = d.set_index(untie_field, ) \
        .drop_duplicates(subset='mdmId', keep='last') \
        .reset_index()
    if 'mdmDeleted' in d.columns:
        d['mdmDeleted'] = d['mdmDeleted'].fillna(False)
        d = d[~d['mdmDeleted']]
    d = d.reset_index(drop=True)
    return d


def drop_duplicated_parquet(d, untie_field='mdmCounterForEntity'):
    """
    Merge updates and delete records from the parquet files in CDS.

    Args:
        d: pd.DataFrame
            Dataframe
        untie_field: str
            Field to be used to untie records with the same `mdmId`. 

    Returns:
        pd.DataFrame

    """

    if untie_field not in d.columns:
        untie_field = 'mdmCounterForEntity'

    d = d.sort_values(untie_field).reset_index(drop=True).drop_duplicates(subset='mdmId', keep='last')
    if 'mdmDeleted' in d.columns:
        d['mdmDeleted'] = d['mdmDeleted'].fillna(False)
        d = d[~d['mdmDeleted']]
    d = d.reset_index(drop=True)
    return d


class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """

    def default(self, obj):
        import numpy as np
        if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
                            np.int16, np.int32, np.int64, np.uint8,
                            np.uint16, np.uint32, np.uint64)):
            return int(obj)
        elif isinstance(obj, (np.float_, np.float16, np.float32,
                              np.float64)):
            return float(obj)
        elif isinstance(obj, (np.ndarray,)):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def _attach_path(branch, trunk):
    '''
    Insert a branch of directories on its trunk.
    '''
    parts = branch.split('/', 1)
    if len(parts) == 1:  # branch is a file
        trunk[_FILE_MARKER].append(parts[0])
    else:
        node, others = parts
        if node not in trunk:
            trunk[node] = defaultdict(dict, ((_FILE_MARKER, []),))
        _attach_path(others, trunk[node])


def prettify_path(d, indent=0):
    '''
    Print the file tree structure with proper indentation.

    :param: d: `dict`
        list of path to prettify
    :param : indent: `int`, defaut `0`
        Ident to use.
    '''
    for key, value in d.items():
        if key == _FILE_MARKER:
            if value:
                print('  ' * indent + str(value))
        else:
            print('  ' * indent + str(key))
            if isinstance(value, dict):
                prettify_path(value, indent + 1)


def ranges(min_v, max_v, nb):
    if min_v == max_v:
        max_v += 1
    step = int((max_v - min_v) / nb) + 1
    step = list(range(min_v, max_v, step))
    if step[-1] != max_v:
        step.append(max_v)
    step = [[step[i], step[i + 1] - 1] for i in range(len(step) - 1)]
    step.append([max_v, None])
    step = [[None, min_v - 1]] + step
    return step


def stream_data(data, step_size, compress_gzip):
    """

    :param data:  `pandas.DataFrame` or `list of dict`,
        Data to be sliced.
    :param step_size: 'int'
        Number of records per slice.
    :param compress_gzip: 'bool'
        If to compress the data to send
    :return: Generator, cont
        Return a slice of `data` and the count of records until that moment.
    """
    if isinstance(data, list):
        is_df = False
    else:
        import pandas as pd
        if isinstance(data, pd.DataFrame):
            is_df = True

    data_size = len(data)
    cont = 0
    for i in range(0, data_size, step_size):
        if is_df:
            data_to_send = data.iloc[i:i + step_size]
            cont += len(data_to_send)
            # print('Sending {}/{}'.format(cont, data_size), end='\r')
            data_to_send = data_to_send.to_json(orient='records', date_format='iso', lines=False)
            if compress_gzip:
                out = io.BytesIO()
                with gzip.GzipFile(fileobj=out, mode="w", compresslevel=9) as f:
                    f.write(data_to_send.encode('utf-8'))
                yield out.getvalue(), cont
            else:
                yield json.loads(data_to_send), cont
        else:
            data_to_send = data[i:i + step_size]
            cont += len(data_to_send)
            # print('Sending {}/{}'.format(cont, data_size), end='\r')
            if compress_gzip:
                out = io.BytesIO()
                with gzip.GzipFile(fileobj=out, mode="w", compresslevel=9) as f:
                    f.write(json.dumps(data_to_send, cls=NumpyEncoder).encode('utf-8'))
                yield out.getvalue(), cont
            else:
                yield data_to_send, cont
    return None, None


class Hashabledict(dict):
    def __hash__(self):
        return hash(frozenset(self))


def zip_folder(path_to_zip):
    path_to_zip = os.path.abspath(path_to_zip)
    path_to_zip = Path(path_to_zip)
    base_name = path_to_zip.name

    zip_file = Path(path_to_zip).with_suffix('.zip')
    with zipfile.ZipFile(zip_file, 'w', zipfile.ZIP_DEFLATED, ) as z:
        for f in list(path_to_zip.rglob('*.*')):
            z.write(f, arcname=f.relative_to(path_to_zip.parents[0]))

    return zip_file




from luigi import BoolParameter
from pycarol.pipeline.task import Task

import os

        
class EasyDockerTask(Task):
    runlocal = BoolParameter(significant=False)  # default is False. So it starts a container by default.

    def _cmd_params(self):
        l = []
        for k, v in self.to_str_params().items():
            if k == 'runlocal':
                continue
            l.append("--{}".format(k).replace("_", "-"))
            l.append(v)
        return l

    @property
    def command(self):
        cmd = ["luigi","--local-scheduler", self.task_family, "--module", self.task_module, "--runlocal",*self._cmd_params() ]
        return cmd

    def get_whole_env(self):
        return {k: v for k, v in os.environ.items()}

    def run(self):
        if self.runlocal:
            print("been there inside a container",self)

            Task.run(self)
        else:
            print("been there before calling docker stuff",self.command)

            docker_environment = self.get_whole_env()
            docker_environment["PYTHONPATH"]="." #TODO isso pode deixar direto na imagem
            print("docker environment\n",docker_environment)

            import docker
            client = docker.from_env()

            client.containers.run(
                self.image,
                self.command,
                environment= docker_environment,
                stream=True,
                # detach=True,
            )
